use std::ops;

pub mod evaluator;
pub mod solver;


fn zero() -> GaussianRational {
    GaussianRational::new(Rational::new(0, 1), Rational::new(0, 1))
}

fn minus_one() -> GaussianRational {
    GaussianRational::new(Rational::new(-1, 1), Rational::new(0, 1))
}

fn one() -> GaussianRational {
    GaussianRational::new(Rational::new(1, 1), Rational::new(0, 1))
}

fn approx_sqrt_f64(n: f64) -> f64 {
    let n = n as f64;
    let mut result = n + 1.;
    let mut factor = 1.;
    while factor > f64::EPSILON {
        while result * result > n {
            result /= 1. + factor;
        }
        result *= 1. + factor;
        factor /= 2.;
    }
    result
}

fn approx_sqrt(n: i128) -> f64 {
    let n = n as f64;
    let mut result = n;
    let mut factor = 1.;
    while factor > f64::EPSILON {
        while result * result > n {
            result /= 1. + factor;
        }
        result *= 1. + factor;
        factor /= 2.;
    }
    result
}

fn get_prime_factors(mut n: i128) -> Vec<i128> {
    if n == 0 {
        return [0].to_vec();
    }
    let mut prime_factors = vec![];
    let mut prime = 3;
    while n != 1 && prime * prime < n + 8 {
        if n % (prime - 1) == 0 {
            prime_factors.push(prime - 1);
            n /= prime - 1;
        } else if n % (prime + 1) == 0 {
            prime_factors.push(prime + 1);
            n /= prime + 1;
        } else {
            if prime > 5 {
                prime += 6;
            } else {
                prime += 1;
            }
        }
    }
    if n != 1 {
        prime_factors.push(n);
    }
    prime_factors
}

fn gcd(a: i128, b: i128) -> i128 {
    if b == 0 {
        a
    } else {
        gcd(b, a % b)
    }
}

#[derive(Debug, PartialEq, Clone, Copy)]
pub struct Rational {
    numerator: i128,
    denominator: i128,
}

impl Rational {
    pub fn new(numerator: i128, denominator: i128) -> Self {
        Self {
            numerator,
            denominator,
        }
    }

    pub fn reduce(&self) -> Self {
        let gcd = gcd(self.numerator, self.denominator);
        let mut res = Rational::new(self.numerator / gcd, self.denominator / gcd);
        if res.denominator < 0 {
            res = Rational::new(-res.numerator, -res.denominator);
        }
        res
    }

    pub fn inverse(&self) -> Self {
        Rational::new(self.denominator, self.numerator)
    }

    pub fn left_shift(&self) -> Self {
        Rational::new(self.numerator, self.denominator * 10)
    }

    fn format_fract(&self) -> String {
        let mut copy = self.clone();
        let factors = get_prime_factors(self.denominator);
        let mut two_count = 0;
        let mut five_count = 0;
        for factor in factors {
            if factor == 2 {
                two_count += 1;
            } else if factor == 5 {
                five_count += 1;
            } else {
                return format!("{}/{}", self.numerator, self.denominator);
            }
        }
        while two_count < five_count {
            copy.numerator *= 2;
            copy.denominator *= 2;
            two_count += 1;
        }
        while five_count < two_count {
            copy.numerator *= 5;
            copy.denominator *= 5;
            five_count += 1;
        }
        if copy.numerator < 0 {
            copy.numerator *= -1;
        }
        format!("{}.{}", copy.numerator / copy.denominator, copy.numerator % copy.denominator)
    }

    pub fn format(&self) -> String {
        match (self.numerator, self.denominator) {
            (_, 1) => format!("{}", self.numerator),
            _ => format!("{}", self.format_fract()),
        }
    }

    pub fn simplify(lhs: Rational, rhs: Rational) -> Self {
        let res = Rational::new(lhs.numerator * rhs.denominator, lhs.denominator * rhs.numerator);
        res.reduce()
    }

    pub fn is_zero(&self) -> bool {
        self.numerator == 0
    }

    pub fn numerator(&self) -> i128 {
        self.numerator
    }

    pub fn denominator(&self) -> i128 {
        self.denominator
    }
}

impl ops::Add<Rational> for Rational {
    type Output = Rational;

    fn add(self, rhs: Rational) -> Rational {
        Rational::new(self.numerator * rhs.denominator + rhs.numerator * self.denominator, self.denominator * rhs.denominator).reduce()
    }
}

impl ops::Add<i128> for Rational {
    type Output = Rational;

    fn add(self, rhs: i128) -> Rational {
        Rational::new(self.numerator + rhs * self.denominator, self.denominator).reduce()
    }
}

impl ops::Add<Rational> for i128 {
    type Output = Rational;

    fn add(self, rhs: Rational) -> Rational {
        Rational::new(self * rhs.denominator + rhs.numerator, rhs.denominator).reduce()
    }
}

impl ops::Sub<Rational> for Rational {
    type Output = Rational;

    fn sub(self, rhs: Rational) -> Rational {
        Rational::new(self.numerator * rhs.denominator - rhs.numerator * self.denominator, self.denominator * rhs.denominator).reduce()
    }
}

impl ops::Sub<i128> for Rational {
    type Output = Rational;

    fn sub(self, rhs: i128) -> Rational {
        Rational::new(self.numerator - rhs * self.denominator, self.denominator).reduce()
    }
}

impl ops::Sub<Rational> for i128 {
    type Output = Rational;

    fn sub(self, rhs: Rational) -> Rational {
        Rational::new(self * rhs.denominator - rhs.numerator, rhs.denominator).reduce()
    }
}

impl ops::Mul<Rational> for Rational {
    type Output = Rational;

    fn mul(self, rhs: Rational) -> Rational {
        Rational::new(self.numerator * rhs.numerator, self.denominator * rhs.denominator).reduce()
    }
}

impl ops::Mul<i128> for Rational {
    type Output = Rational;

    fn mul(self, rhs: i128) -> Rational {
        Rational::new(self.numerator * rhs, self.denominator).reduce()
    }
}

impl ops::Mul<Rational> for i128 {
    type Output = Rational;

    fn mul(self, rhs: Rational) -> Rational {
        Rational::new(self * rhs.numerator, rhs.denominator).reduce()
    }
}

impl ops::Div<Rational> for Rational {
    type Output = Rational;

    fn div(self, rhs: Rational) -> Rational {
        self * rhs.inverse()
    }
}

impl ops::Div<i128> for Rational {
    type Output = Rational;

    fn div(self, rhs: i128) -> Rational {
        let rhs_inverse = Rational::new(1, rhs);
        self * rhs_inverse
    }
}

impl ops::Div<Rational> for i128 {
    type Output = Rational;

    fn div(self, rhs: Rational) -> Rational {
        self * rhs.inverse()
    }
}

#[derive(Debug, PartialEq, Clone, Copy)]
pub struct GaussianRational {
    real: Rational,
    imaginary: Rational,
}

impl GaussianRational {
    pub fn new(real: Rational, imaginary: Rational) -> Self {
        Self {
            real,
            imaginary,
        }
    }

    pub fn conjugate(&self) -> Self {
        GaussianRational::new(self.real, -1 * self.imaginary)
    }

    pub fn format(&self, parenthesis: bool) -> String {
        match (self.real().is_zero(), self.imaginary.is_zero(), parenthesis) {
            (true, true, _) => format!(""),
            (true, false, _) => format!("{}i", self.imaginary.format()),
            (false, true, _) => format!("{}", self.real.format()),
            (false, false, false) => format!("{} + {}i", self.real.format(), self.imaginary.format()),
            (false, false, true) => format!("({} + {}i)", self.real.format(), self.imaginary.format())
        }
    }

    pub fn is_real(&self) -> bool {
        self.imaginary.is_zero()
    }

    pub fn real(&self) -> Rational {
        self.real
    }

    pub fn imaginary(&self) -> Rational {
        self.imaginary
    }
}

impl ops::Add<GaussianRational> for GaussianRational {
    type Output = GaussianRational;

    fn add(self, rhs: GaussianRational) -> GaussianRational {
        GaussianRational::new(self.real + rhs.real, self.imaginary + rhs.imaginary)
    }
}

impl ops::Sub<GaussianRational> for GaussianRational {
    type Output = GaussianRational;

    fn sub(self, rhs: GaussianRational) -> GaussianRational {
        GaussianRational::new(self.real - rhs.real, self.imaginary - rhs.imaginary)
    }
}

impl ops::Mul<GaussianRational> for GaussianRational {
    type Output = GaussianRational;

    fn mul(self, rhs: GaussianRational) -> GaussianRational {
        GaussianRational::new(self.real * rhs.real - self.imaginary * rhs.imaginary, self.real * rhs.imaginary + self.imaginary * rhs.real)
    }
}

impl ops::Mul<i128> for GaussianRational {
    type Output = GaussianRational;

    fn mul(self, rhs: i128) -> GaussianRational {
        GaussianRational::new(self.real * rhs, self.imaginary * rhs)
    }
}

impl ops::Mul<GaussianRational> for i128 {
    type Output = GaussianRational;

    fn mul(self, rhs: GaussianRational) -> GaussianRational {
        GaussianRational::new(self * rhs.real, self * rhs.imaginary)
    }
}

impl ops::Div<GaussianRational> for GaussianRational {
    type Output = GaussianRational;

    fn div(self, rhs: GaussianRational) -> GaussianRational {
        let cd_squared = rhs.real * rhs.real + rhs.imaginary * rhs.imaginary;
        GaussianRational::new(Rational::simplify(self.real * rhs.real + self.imaginary * rhs.imaginary, cd_squared), Rational::simplify(self.imaginary * rhs.real - self.real * rhs.imaginary, cd_squared))
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    
    #[test]
    fn constant_generators() {
        assert_eq!(zero(), GaussianRational::new(Rational::new(0, 1), Rational::new(0, 1)));
        assert_eq!(minus_one(), GaussianRational::new(Rational::new(-1, 1), Rational::new(0, 1)));
        assert_eq!(one(), GaussianRational::new(Rational::new(1, 1), Rational::new(0, 1)));
    }

    #[test]
    fn get_prime_factors_test() {
        assert_eq!(get_prime_factors(5), vec![5]);
        assert_eq!(get_prime_factors(8), vec![2, 2, 2]);
        assert_eq!(get_prime_factors(4), vec![2, 2]);
        assert_eq!(get_prime_factors(9), vec![3, 3]);
        assert_eq!(get_prime_factors(10), vec![2, 5]);
        assert_eq!(get_prime_factors(18), vec![2, 3, 3]);
        assert_eq!(get_prime_factors(16), vec![2, 2, 2, 2]);
    }

    #[test]
    fn gcd_test() {
        assert_eq!(gcd(7, 3), 1);
        assert_eq!(gcd(7, -3), 1);
        assert_eq!(gcd(-7, 3), -1);
        assert_eq!(gcd(-7, -3), -1);


        assert_eq!(gcd(25, 20), 5);
        assert_eq!(gcd(25, -20), 5);
        assert_eq!(gcd(-25, 20), -5);
        assert_eq!(gcd(-25, -20), -5);

        assert_eq!(gcd(30, 30), 30);
    }

    mod rational {
        use super::*;

        #[test]
        fn new() {
            let rational = Rational::new(5, 10);
            assert_eq!(rational, Rational { numerator: 5, denominator: 10 });
        }

        #[test]
        fn reduce() {
            let rational = Rational::new(20, 10);
            assert_eq!(rational.reduce(), Rational::new(2, 1));

            let rational = Rational::new(7, -3);
            assert_eq!(rational.reduce(), Rational::new(-7, 3));

        }

        #[test]
        fn inverse() {
            let rational = Rational::new(5, 2);
            assert_eq!(rational.inverse(), Rational::new(2, 5));
        }

        #[test]
        fn left_shift() {
            let rational = Rational::new(7, 3);
            assert_eq!(rational.left_shift(), Rational::new(7, 30));
        }

        #[test]
        fn simplify() {
            let a = Rational::new(10, 3);
            let b = Rational::new(7, 2);

            assert_eq!(Rational::simplify(a, b), Rational::new(20, 21));


            let a = Rational::new(10, 5);
            let b = Rational::new(5, 2);

            assert_eq!(Rational::simplify(a, b), Rational::new(4, 5));
        }

        #[test]
        fn is_zero() {
            let rational = Rational::new(0, 4);
            assert_eq!(rational.is_zero(), true);

            let rational = Rational::new(3, 4);
            assert_eq!(rational.is_zero(), false);
        }

        #[test]
        fn format() {
            let whole = Rational::new(5, 1);
            let rational = Rational::new(5, 3);
            let floating = Rational::new(5, 4);
            assert_eq!(whole.format(), String::from("5"));
            assert_eq!(rational.format(), String::from("5/3"));
            assert_eq!(floating.format(), String::from("1.25"));
            let floating = Rational::new(37, 25);
            assert_eq!(floating.format(), String::from("1.48"));
        }

        #[test]
        fn getters() {
            let rational = Rational::new(5, 7);
            assert_eq!(rational.numerator(), 5);
            assert_eq!(rational.denominator(), 7);
        }

        #[test]
        fn ops() {
            let a = Rational::new(5, 7);
            let b = Rational::new(12, 14);
            let n = 37;

            assert_eq!(a + b, Rational::new(11, 7));
            assert_eq!(a - b, Rational::new(-1, 7));
            assert_eq!(a * b, Rational::new(30, 49));
            assert_eq!(a / b, Rational::new(5, 6));

            assert_eq!(a + n, Rational::new(264, 7));
            assert_eq!(n + a, Rational::new(264, 7));
            assert_eq!(a - n, Rational::new(-254, 7));
            assert_eq!(n - a, Rational::new(254, 7));
            assert_eq!(a * n, Rational::new(185, 7));
            assert_eq!(n * a, Rational::new(185, 7));
            assert_eq!(a / n, Rational::new(5, 259));
            assert_eq!(n / a, Rational::new(259, 5));
        }
    }

    mod gaussian_rational {
        use super::*;

        #[test]
        fn new() {
            let gaussian_rational = GaussianRational::new(Rational::new(5, 7), Rational::new(8, 3));
            assert_eq!(gaussian_rational, GaussianRational {
                real: Rational::new(5, 7),
                imaginary: Rational::new(8, 3)
            });
        }

        #[test]
        fn conjugate() {
            let gaussian_rational = GaussianRational::new(Rational::new(5, 7), Rational::new(8, 3));
            assert_eq!(gaussian_rational.conjugate(), GaussianRational::new(Rational::new(5, 7), Rational::new(-8, 3)));
        }

        #[test]
        fn is_real() {
            let gaussian_rational = GaussianRational::new(Rational::new(5, 7), Rational::new(8, 3));
            assert_eq!(gaussian_rational.is_real(), false);

            let gaussian_rational = GaussianRational::new(Rational::new(5, 7), Rational::new(0, 3));
            assert_eq!(gaussian_rational.is_real(), true);
        }

        #[test]
        fn format() {
            let whole = Rational::new(5, 1);
            let rational = Rational::new(5, 3);
            let zero = Rational::new(0, 1);
            let a = GaussianRational::new(zero, zero);
            let b = GaussianRational::new(zero, rational);
            let c = GaussianRational::new(whole, zero);
            let d = GaussianRational::new(whole, rational);
            assert_eq!(a.format(false), String::from(""));
            assert_eq!(b.format(false), String::from("5/3i"));
            assert_eq!(c.format(false), String::from("5"));
            assert_eq!(d.format(false), String::from("5 + 5/3i"));
        }

        #[test]
        fn getters() {
            let gaussian_rational = GaussianRational::new(Rational::new(5, 7), Rational::new(8, 3));
            assert_eq!(gaussian_rational.real(), Rational::new(5, 7));
            assert_eq!(gaussian_rational.imaginary(), Rational::new(8, 3));
        }

        #[test]
        fn ops() {
            let a = GaussianRational::new(Rational::new(5, 7), Rational::new(8, 3));
            let b = GaussianRational::new(Rational::new(8, 7), Rational::new(7, 3));
            let n = 3;

            assert_eq!(a + b, GaussianRational::new(Rational::new(13, 7), Rational::new(5, 1)));
            assert_eq!(a - b, GaussianRational::new(Rational::new(-3, 7), Rational::new(1, 3)));
            assert_eq!(a * b, GaussianRational::new(Rational::new(-2384, 441), Rational::new(33, 7)));
            assert_eq!(a / b, GaussianRational::new(Rational::new(3104, 2977), Rational::new(609, 2977)));

            assert_eq!(a * n, GaussianRational::new(Rational::new(15, 7), Rational::new(8, 1)));
            assert_eq!(n * a, GaussianRational::new(Rational::new(15, 7), Rational::new(8, 1)));
        }
    }
}